/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * History:
 * 31-Oct-01 fhk; modified to support ldap check via bind on dn of user.
 * --------------------------------------------------------------------------*/

#include "jsm.h"

#include <lber.h>
#include <ldap.h>

#define TERROR_LDAPAUTH          (terror) {450,"Not Member"}

#define LDAPCHK_SUCCESS 1 /* TRUE */ 
#define LDAPCHK_FAIL    0 /* FALSE */

/* -------------------------------------------------------- */
/* Adjust these setting for your own LDAP server            */
/* these are loaded using the configurations in jabber.xml. */
/* -------------------------------------------------------- */
char HOSTNAME[50]; // the ldap server host name.
int PORT_NUMBER;   // the ldap port number.
char BASE_DN[64];  // the base dn for ldap searches.
char UIDATTR[50];  // the attribute that is used to map jabber login name to dn.

/* 
	Check ldap attribute.
		Connects to membership directory, finds t and checks the value of one attribute.
*/
short checkLdapAttr(short *cause, char *filter, char *ctpasswd)
{
   short ret = LDAPCHK_FAIL;
   LDAP *ld;
   LDAPMessage *result, *e;
   char **vals;
   int rc;
   char *attrs[] = {"userpassword", NULL}; // attributes to return
   // int res; // used on OLD way.
   // char *upasswd;  // password returned by ldap // used on OLD way.
   char *dn = NULL;
   int version;

   *cause = 0;

   /* Get a handle to an LDAP connection. */
   if ((ld = ldap_init(HOSTNAME, PORT_NUMBER)) == NULL) 
   {  
      *cause = 1;
      log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 1 (ldap connect failed)");
      return LDAPCHK_FAIL;
   }
   version = LDAP_VERSION3;
   ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &version);

   /* Bind anonymously to the LDAP server. */
   if ((rc = ldap_simple_bind_s(ld, NULL, NULL)) != LDAP_SUCCESS) 
   {
      *cause = 2;
      log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 2 (ldap bind failed) on server %s, port %d",HOSTNAME,PORT_NUMBER);
      return LDAPCHK_FAIL;
   }

   log_debug(ZONE,"LDAP bind on server %s, port %d successful",HOSTNAME,PORT_NUMBER);
   
   /* Search for the entry. */
   log_debug(ZONE,"LDAP search ldap_search_s((ld, %s, LDAP_SCOPE_SUBTREE, %s, 'userpassword', LDAP_NO_LIMIT, &result)",BASE_DN,filter);
   if ((rc = ldap_search_s(ld, BASE_DN, LDAP_SCOPE_SUBTREE, filter, attrs, LDAP_NO_LIMIT, &result)) != LDAP_SUCCESS) 
   {
      *cause = 3;
      log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 3 (ldap search failed)");
      return LDAPCHK_FAIL;   
   } 

   /* there should be only one matching entry, use the first entry in all cases */
   e = ldap_first_entry(ld, result); 
   if (e != NULL) 
   {
     /* Get the attributes  */
     dn = ldap_get_dn(ld, e); // the DN
     log_debug(ZONE,"attempting login as dn=%s for %s using password %s",dn,filter,ctpasswd);
     
     if ((vals = ldap_get_values(ld, e, "userpassword")) != NULL) 
     {
        /* old way,
        upasswd = vals[0];
        res = strncmp(upasswd, ctpasswd, strlen(ctpasswd));
        if (res == 0 && (strlen(upasswd) == strlen(ctpasswd))){
         log_debug(ZONE,"upasswd=%s len=%d  returning LDAPCHK_SUCCESS",upasswd, strlen(ctpasswd));
         ret = LDAPCHK_SUCCESS;
        }
        */

        if ((rc = ldap_simple_bind_s(ld, dn, ctpasswd)) == LDAP_SUCCESS) {
         log_debug(ZONE,"returning LDAPCHK_SUCCESS for dn %s",dn);
         ret = LDAPCHK_SUCCESS;
        }else
        {
          *cause = 10;   
          log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 10 (invalid password)");
          ret = LDAPCHK_FAIL;
        }
     }else 
     {
        *cause = 4;
        log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 4 (no such attribute, 'userpassword') for dn: %s",dn);
        ret = LDAPCHK_FAIL;
     } 

     ldap_value_free(vals);  
     ldap_memfree(dn);
     ldap_unbind(ld);

   }else
   {
     *cause = 11;
     log_debug(ZONE,"returning LDAPCHK_FAIL, cause set to 11 (user not found)");
     ret = LDAPCHK_FAIL;
   }

   ldap_msgfree(result);

   return ret;
}

/*
	Login Authentication. 
		Calls the ldap control routine to check username/password on ldap.
*/
mreturn mod_auth_ldap_auth(mapi m, void *arg)
{
   short cause;
   char *in_pw = NULL;
   char *in_un = NULL;
   char *filter = NULL;
   
   log_debug(ZONE,"start of auth control cleartext");

   if (m->packet->type != JPACKET_IQ)return M_IGNORE;

   log_debug(ZONE,"subtype=%d",jpacket_subtype(m->packet));

   if(jpacket_subtype(m->packet) == JPACKET__GET)
   { /* type=get means we flag that the server can do digest auth */
      xmlnode_insert_tag(m->packet->iq,"password");
      log_debug(ZONE,"type is JPACKET__GET, inserted tag 'password', returning M_PASS");
      return M_PASS;
   }

   if (jpacket_subtype(m->packet) != JPACKET__SET){
    log_debug(ZONE,"type is JPACKET__SET, returning M_PASS");
    return M_PASS;
   }
           
   /* check the packet */ 
   log_debug(ZONE,"what does my packet look like: %s",xmlnode2str(m->packet->x));

   if ((in_un = xmlnode_get_tag_data(m->packet->iq,"username")) == NULL) 
   {
           jutil_error(m->packet->x, TERROR_LDAPAUTH);
           log_debug(ZONE,"username missing, returning M_HANDLED");
           return M_HANDLED;
   }
           
   if ((in_pw = xmlnode_get_tag_data(m->packet->iq,"password")) == NULL) 
   {
           jutil_error(m->packet->x, TERROR_LDAPAUTH);
           log_debug(ZONE,"password missing, returning M_HANDLED");
           return M_HANDLED;               
   }

   filter = (char *) malloc(sizeof(char)*(strlen(in_un)+sizeof(char)*(strlen(UIDATTR))+4));
   sprintf(filter,"%s=%s",UIDATTR,in_un);

   log_debug(ZONE,"user=%s password=%s filter=%s",in_un,in_pw,filter);

   if (checkLdapAttr(&cause, filter, in_pw) == LDAPCHK_SUCCESS)
   {
           log_debug(ZONE,"LDAP OK cause=%d", cause);
           jutil_iqresult(m->packet->x);
   }
   else
   {       
           log_debug(ZONE,"LDAP NOT FOUND cause=%d", cause);
           jutil_error(m->packet->x, TERROR_LDAPAUTH);     
   }

   free(filter);
   log_debug(ZONE,"returning M_HANDLED");
   return M_HANDLED;   
}

/*
	Block all Registration.
	   Used to Forbid Registering a new user.
		Normally we use this callback to register a new user. But being an ldap auth.
		I am assuming that we are registering users somewhere else, and there is no
		registration through jabber server.
*/
mreturn mod_auth_ldap_noreg(mapi m, void *arg)
{
   
	if (m->packet->type != JPACKET_IQ)return M_IGNORE;
	
	log_debug(ZONE,"no registration");
		
	if (jpacket_subtype(m->packet) != JPACKET__SET){
	 log_debug(ZONE,"jpacket_subtype is not JPACKET__SET, returning M_PASS");
         return M_PASS;
        }

	jutil_error(m->packet->x, TERROR_FORBIDDEN);
	
	log_debug(ZONE,"returning M_HANDLED");
	return M_HANDLED;
}

/* Check ldap and Register
		Used to check membership before registering the user on jabber server 
*/
mreturn mod_auth_ldap_reg(mapi m, void *arg)
{
   short cause;
	char *in_pw = NULL;
	char *in_un = NULL;
	char *filter = NULL;
	
	if (m->packet->type != JPACKET_IQ)return M_IGNORE;
	
	log_debug(ZONE,"registration check");
		
	if (jpacket_subtype(m->packet) != JPACKET__SET) return M_PASS;

	if ((in_un = xmlnode_get_tag_data(m->packet->iq,"username")) == NULL) 
	{
		jutil_error(m->packet->x, TERROR_LDAPAUTH);
                log_debug(ZONE,"username missing, returning M_HANDLED");
		return M_HANDLED;
	}
	
	if ((in_pw = xmlnode_get_tag_data(m->packet->iq,"password")) == NULL) 
	{
		jutil_error(m->packet->x, TERROR_LDAPAUTH);
                log_debug(ZONE,"password missing, returning M_HANDLED");
		return M_HANDLED;		
	}
	
        filter = (char *) malloc(sizeof(char)*(strlen(in_un)+sizeof(char)*(strlen(UIDATTR))+4));
        sprintf(filter,"%s=%s",UIDATTR,in_un);

        log_debug(ZONE,"user=%s password=%s filter=%s",in_un,in_pw,filter);

	if (checkLdapAttr(&cause, filter, in_pw) == LDAPCHK_SUCCESS)
	{
        	log_debug(ZONE,"Found cause, cause=%d, returning M_PASS",cause);
		return M_PASS;
	}
	else
	{	
        	log_debug(ZONE,"Not Found, cause=%d, returning M_HANDLED",cause);
		jutil_error(m->packet->x, TERROR_LDAPAUTH);
		return M_HANDLED;
	}	
}

/* the main startup/initialization function */
void mod_auth_ldap_z(jsmi si)
{	
	xmlnode myx;
	char *pNode;

	if ((myx = js_config(si, "ldapchk")) != NULL) 
	{
		log_debug(ZONE,"ldap settings:\n%s",xmlnode2str(myx));
		
		pNode = xmlnode_get_tag_data(myx,"host");
		strcpy(HOSTNAME, pNode);
		log_debug(ZONE,"hostname=%s",HOSTNAME);
		
		pNode = xmlnode_get_tag_data(myx,"port");
		PORT_NUMBER = atoi(pNode);
		log_debug(ZONE,"port=%d",PORT_NUMBER);
		
		pNode = xmlnode_get_tag_data(myx,"basedn");
		strcpy(BASE_DN, pNode);
		log_debug(ZONE,"basedn=%s",BASE_DN);

		pNode = xmlnode_get_tag_data(myx,"uidattr");
		strcpy(UIDATTR, pNode);
		log_debug(ZONE,"uidattr=%s",UIDATTR);
	}

	// use mod_auth_ldap_noreg for blocking registration
	log_debug(ZONE,"calling js_mapi_register, e_REGISTER, function: mod_auth_ldap_reg");
	js_mapi_register(si, e_REGISTER, mod_auth_ldap_reg, NULL);
	log_debug(ZONE,"calling js_mapi_register, e_AUTH, function: mod_auth_ldap_auth");
        js_mapi_register(si, e_AUTH, mod_auth_ldap_auth, NULL); 
}
